# -----------------------------------------
# Endroid - XMPP Bot
# Copyright 2012, Ensoft Ltd.
# Created by Jonathan Millican
# -----------------------------------------

import sqlite3
import os.path
import logging

__all__ = (
    'Value', 'LIKE',
    'Field',
    'Database',
    'EndroidUniqueID', 'EndroidLastModified',
)


class TableRow(dict):
    """
    A dict with special properties for access the magic fields of the row.

    These should not be instantiated directly, but are instead returned from
    "fetch" operations on tables.

    """
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(TableRow, self).__init__(*args, **kwargs)

    @property
    def id(self):
        return self[EndroidUniqueID]

    @property
    def last_modified(self):
        return self[EndroidLastModified]


class Value(object):
    """
    Wrapper for a value to be matched in a condition.

    'cmp' must be one of the comparison operators supported by SQL.

    """
    def __init__(self, val, cmp='='):
        self.value = str(val)
        self._cmp = str(cmp)

    def __str__(self):
        return self.value

    @classmethod
    def as_condition(cls, obj):
        if isinstance(obj, cls):
            return obj._cmp + '?'
        else:
            return Value.as_condition(Value(obj))


class LIKE(Value):
    "Value subclass for handling SQl 'LIKE' comparisons."
    def __init__(self, val):
        super(LIKE, self).__init__("%{}%".format(val), "LIKE")


class Field(object):
    """
    Class representing a Field of a Table.

    These need usually only be instantiated when a table is first being
    created. Note that raw strings can be used as fields in create calls,
    and they simply get the default type.

    """
    def __init__(self, name, type=None, default=None):
        self._name = Field.as_str(name)
        self._type = type
        self._default = default

    def __str__(self):
        return self._name

    @classmethod
    def as_create_str(cls, obj):
        if isinstance(obj, cls):
            return "{} {} {}".format(obj._name, obj._type if obj._type else "",
                             "default " + obj._default if obj._default else "")
        else:
            return Field.as_create_str(Field(obj))

    @classmethod
    def as_str(cls, obj):
        if isinstance(obj, cls):
            return obj._name
        else:
            return Database._sanitize(obj)


class Table(object):
    """
    Wrapper around a table (so that the table name doesn't need to be
    passed as the first argument all the time).

    """
    def __init__(self, name, db):
        self._name = name
        self._db = db

    def create(self, fields):
        return self._db.create_table(self._name, fields)

    def exists(self):
        return self._db.table_exists(self._name)

    def fetch(self, fields, conditions={}, distinct=False, max=False):
        return self._db.fetch(self._name, fields, conditions, distinct, max)

    def count(self, conditions):
        return self._db.count(self._name, conditions)

    def insert(self, fields):
        return self._db.insert(self._name, fields)

    def delete(self, conditions={}):
        return self._db.delete(self._name, conditions)

    def update(self, fields, conditions):
        return self._db.update(self._name, fields, conditions)

    def empty(self):
        return self._db.empty_table(self._name)

    def drop(self):
        return self._db.delete_table(self._name)

    @property
    def name(self):
        return self._db._tName(self._name)


class Database(object):
    """
    Wrapper round an sqlite3 Database.

    All accesses are synchronous, TODO use twisted.enterprise.adbapi to 
    asynchronise them.

    """
    connection = None
    cursor = None
    file_name = None
    
    @staticmethod
    def setFile(file_name):
        Database.file_name = os.path.expanduser(file_name)
    
    def __init__(self, modName):
        if Database.connection == None:
            Database.connection = sqlite3.connect(Database.file_name)
            Database.cursor = Database.connection.cursor()
        self.modName = modName
    
    def _tName(self, name):
        return self.modName + "_" + name
    
    @staticmethod
    def _sanitize(inp):
        inp = inp.replace('\\', '\\\\')
        inp = inp.replace('"', '\\"')
        return '"' + inp + '"'
    
    @staticmethod
    def _qMarks(fields, delim=""):
        return ", ".join(delim + "?" + delim for f in fields)
    
    @staticmethod
    def _stringFromFieldNames(fields):
        return ", ".join(Field.as_str(f) for f in fields)

    @staticmethod
    def _stringFromListItems(list):
        return ", ".join(Field.as_str(l) for l in list)
    
    @staticmethod
    def _tupleFromFieldValues(fields):
        return tuple(str(v) for v in fields.values())
    
    @staticmethod
    def _buildConditions(conditions):
        return " and ".join(Field.as_str(f) + Value.as_condition(v)
                            for f, v in conditions.items()) or "1"

    def table(self, name):
        return Table(name, self)

    def create_table(self, name, fields):
        """
        Create a new table in the database called 'name' and containing fields 
        'fields' (an iterable of strings giving field titles).

        """
        if any(str(f).startswith('_endroid') for f in fields):
            raise ValueError("An attempt was made to create a table with "
                             "system-reserved column-name (prefix '_endroid').")
        n = Database._sanitize(self._tName(name))
        fields_string = ', '.join(map(Field.as_create_str,
                                  [EndroidUniqueID] + list(fields)))
        query = 'CREATE TABLE {0} ({1});'.format(n, fields_string)
        Database.raw(query)

        if EndroidLastModified in fields:
            Database.raw("CREATE TRIGGER {3}_last_modified AFTER UPDATE ON {0} "
                         "BEGIN UPDATE {0} SET {1} = DATETIME('now') "
                         "WHERE {2} = new.{2}; END;"
                         .format(n, EndroidLastModified, EndroidUniqueID,
                                 self._tName(name)))
    
    def table_exists(self, name):
        """Check to see if a table called 'name' exists in the database."""
        n = Database._sanitize(self._tName(name))
        query = "SELECT `name` FROM `sqlite_master` WHERE `type`='table' AND `name`={0};".format(n)
        Database.raw(query)
        r = Database.cursor.fetchall()
        count = len(r)
        return count != 0
    
    def insert(self, name, fields):
        """
        Insert a row into table 'name'.
        
        Fields is a dictionary mapping field names (as defined in 
        create_table) to values.

        """
        n = Database._sanitize(self._tName(name))
        query = "INSERT INTO {0} ({1}) VALUES ({2});".format(n,
                               Database._stringFromFieldNames(fields),
                               Database._qMarks(fields, ''))
        tup = Database._tupleFromFieldValues(fields)
        Database.raw(query, tup)
        return Database.cursor.lastrowid
    
    def fetch(self, name, fields, conditions={}, distinct=False, max=False):
        """
        Get data from the table 'name'. 

        Returns a list of dictionaries mapping 'fields' to their values, one 
        dictionary for each row which satisfies a condition in conditions.

        Conditions is a dictionary mapping field names to values. A result
        will only be returned from a row if its values match those in conditions

        E.g.: conditions = {'user': 'JoeBloggs'} 
        will match only fields in rows which have JoeBloggs in the 'user' field.

        """
        n = Database._sanitize(self._tName(name))
        if not distinct and not max:
            fields = list(fields) + [EndroidUniqueID]
        if max:
            cols = "MAX({})".format(Database._stringFromListItems(fields))
        else:
            cols = Database._stringFromListItems(fields)
        query = "SELECT {3}{0} FROM {1} WHERE ({2});".format(
            cols, n, Database._buildConditions(conditions),
            "DISTINCT " if distinct else "")
        Database.raw(query, Database._tupleFromFieldValues(conditions))
        c = Database.cursor.fetchall()
        rows = [TableRow(dict(zip(fields, item))) for item in c]
        return rows

    def count(self, name, conditions):
        "Return the number of rows in table 'name' which satisfy conditions."
        n = Database._sanitize(self._tName(name))
        query = "SELECT COUNT(*) FROM {0} WHERE ({1});".format(n, Database._buildConditions(conditions))
        r = Database.raw(query, Database._tupleFromFieldValues(conditions)).fetchall()
        return r[0][0]

    def delete(self, name, conditions={}):
        """Delete rows from table 'name' which satisfy conditions."""
        n = Database._sanitize(self._tName(name))
        query = "DELETE FROM {0} WHERE ({1});".format(n, Database._buildConditions(conditions))
        Database.raw(query, Database._tupleFromFieldValues(conditions))
        return Database.cursor.rowcount
    
    def update(self, name, fields, conditions):
        """
        Update rows in table 'name' which satisfy conditions. 

        Fields is a dictionary mapping the field names to their new values.

        """
        n = Database._sanitize(self._tName(name))
        query = "UPDATE {0} SET {1} WHERE ({2});".format(n, Database._buildConditions(fields), Database._buildConditions(conditions))
        tup = Database._tupleFromFieldValues(fields)
        tup = tup + Database._tupleFromFieldValues(conditions)
        Database.raw(query, tup)
        return Database.cursor.rowcount

    def empty_table(self, name):
        """Remove all rows from table 'name'."""
        return self.delete(name)
    
    def delete_table(self, name):
        """Delete table 'name'."""
        n = Database._sanitize(self._tName(name))
        query = "DROP TABLE {0};".format(n)
        Database.raw(query)

    @staticmethod
    def raw(command, params=()):
        logging.debug("SQL command {} params {}".format(command, params))
        p = Database.cursor.execute(command, params)
        Database.connection.commit()
        return p

# Export constants for system column names
EndroidUniqueID = Field('_endroid_unique_id',
                        type="INTEGER PRIMARY KEY AUTOINCREMENT")
EndroidLastModified = Field('_endroid_last_modified', type="DATETIME",
                            default="current_timestamp")
