# -----------------------------------------------------------------------------
# EnDroid - Remote notification plugin
# Copyright 2012, Ensoft Ltd
# -----------------------------------------------------------------------------

import os
import re
import json
import time
import shlex
import logging
import cfgparser

from Crypto.PublicKey import RSA
from twisted.conch.ssh import keys
from twisted.internet import reactor, defer, threads
from twisted.internet.protocol import ProcessProtocol

from endroid.database import Database, EndroidUniqueID, EndroidLastModified
from endroid.usermanagement import JID
from endroid.plugins.command import CommandPlugin, command

from .web import KermitRoot
from . import ssh


class Kermit(CommandPlugin):
    dependencies = ('endroid.plugins.httpinterface',)
    help = "Automatically commit patches without having to do the legwork."
    DB_TABLE = "Queue"
    KEY_TABLE = "Keys"

    def endroid_init(self):
        self.http = self.get('endroid.plugins.httpinterface')
        self.http.register_resource(self, KermitRoot(self))

        self.config_dir = self.vars.get("config_dir")

        self.db = self.database.table(self.DB_TABLE)
        self.keydb = self.database.table(self.KEY_TABLE)

        if not self.db.exists():
            self.db.create(("user", "target", "status", EndroidLastModified))
        if not self.keydb.exists():
            self.keydb.create(("user", "keystr"))

        self.waiters = []


    def notify_waiters(self):
        for querier, waiter in self.waiters:
            reactor.callLater(1, querier.respond, waiter)
        del self.waiters[:]


    def read_config(self, target=None):
        files = [os.path.join(self.config_dir, "kermit.cfg")]
        if target:
            files.append(os.path.join(self.config_dir, target + ".cfg"))
        c = cfgparser.CfgParser()
        c.read(files)
        return c


    # Key Management

    def _generate_key(self, username):
        key = RSA.generate(2048)
        return keys.Key.fromString(key.exportKey("PEM"))

    def generate_key(self, username):
        return threads.deferToThread(self._generate_key, username)

    def add_key(self, username, keystr):
        key = keys.Key.fromString(keystr)
        if key.isPublic():
            raise ValueError("Uploaded key must be a private key")

        id = self.keydb.insert({
          "user": username,
          "keystr": key.toString("OPENSSH"),
        })
        return {
            'idx': id,
            'type': key.public().type(),
            'keystr': key.public().toString("OPENSSH"),
        }


    def remove_key(self, username, idx):
        self.keydb.delete({"user": username, EndroidUniqueID: idx})


    def get_private_keys(self, username):
        rows = self.keydb.fetch(("keystr",), {"user": username})
        return [keys.Key.fromString(k['keystr']) for k in rows]


    def list_keys(self, username):
        rows = self.keydb.fetch(("keystr",), {"user": username})
        ks = ((k.id, keys.Key.fromString(k['keystr']).public()) for k in rows)
        return [{'idx': i, 'type': k.type(), 'keystr': k.toString("OPENSSH")}
                for i, k in ks]


    @staticmethod
    def _parse(arg):
        parts = shlex.split(arg)

        if len(parts) != 5:
            raise ValueError("Invalid commit command")
        if parts[1].lower() != 'to' or parts[3].lower() != 'against':
            raise ValueError("Malformed commit command")

        return parts[0::2]


    def abort_commit(self, deferred, key):
        row = self.db.fetch(("target", "user", "status"),
                             {EndroidUniqueID: key})
        if not row:
            raise ValueError("Unknown queue item ({})".format(key))
        target = row[0]['target']
        user = row[0]['user']
        status = json.loads(row[0]['status'])

        config = self.read_config(target)
        config.set("DEFAULT", "target", target)
        config.set("DEFAULT", "username", user)
        config.set("DEFAULT", "key", "kermit-{}".format(key))

        def _reset_status(_):
            status['currstatus']['running'] = []
            status['currstatus']['failed'] = []
            status['currstatus']['failed_at'] = {}
            status['aborted_at'] = time.time()
            self.db.update({
                "status": json.dumps(status),
            }, {EndroidUniqueID: key})
            self.notify_waiters()
            return key

        d = ssh.spawn_process(config.get(target, "server", default="localhost"),
                              config.getlist(target, "fingerprints",
                                             default=[]),
                              user, self.get_private_keys(user),
                              "-a kermit-{}".format(key),
                              config)

        d.addCallback(_reset_status)
        d.chainDeferred(deferred)


    def resubmit_commit(self, deferred, key):
        row = self.db.fetch(("status",), {EndroidUniqueID: key})
        if not row:
            raise ValueError("Unknown queue item ({})".format(key))
        args = json.loads(row[0]['status'])['inputargs']

        username = args['username']
        commitdiff = args['commitdiff']
        target = args['target']
        buglist = args['buglist']
        commitmsg = args['commitmsg']
        stages = args.get('stages', '')
        ws_disable = args.get('ws_disable', '')
        ws_reset = args.get('ws_reset', '')

        self.submit_commit(deferred, username, commitdiff, target, buglist,
                           commitmsg, stages, key=key, ws_disable=ws_disable,
                           ws_reset=ws_reset)


    def submit_commit(self, deferred, user, commitdiff, target, buglist,
                      commitmsg='', stages='', key=None, ws_disable='',
                      ws_reset=''):
        if not os.path.exists(os.path.join(self.config_dir, target + '.cfg')):
            deferred.errback(OSError("No configuration file for target"))
            return
        config = self.read_config(target)
        config.set("DEFAULT", "commitdiff", commitdiff)
        config.set("DEFAULT", "target", target)
        config.set("DEFAULT", "buglist", buglist)
        config.set("DEFAULT", "commitmsg", commitmsg)
        config.set("DEFAULT", "username", user)
        config.set("DEFAULT", "stages", stages)
        config.set("DEFAULT", "ws_disable", ws_disable)
        config.set("DEFAULT", "ws_reset", ws_reset)

        def _reset_status(_):
            rows = self.db.fetch(("status",), {EndroidUniqueID: key})
            if rows:
                status = json.loads(rows[0]['status'])
                status['currstatus']['running'] = []
                status['currstatus']['failed'] = []
                status['currstatus']['failed_at'] = {}
                status['aborted_at'] = 0
                self.db.update({
                    "status": json.dumps(status),
                }, {EndroidUniqueID: key})
                self.notify_waiters()
            return _

        deferred.addCallback(_reset_status)

        args = {
            'username': user,
            'commitdiff': commitdiff,
            'target': target,
            'buglist': buglist,
            'commitmsg': commitmsg,
            'stages': stages,
            'ws_disable': ws_disable,
            'ws_reset': ws_reset,
        }

        if key:
            rowid = key
        else:
            rowid = self.db.insert({
                "user": user,
                "target": target,
                "status": json.dumps({
                    'inputargs': args,
                    'currstatus': {},
                    'aborted_at': 0,
                }),
            })
            def _cleanup(fail):
                self.db.delete({EndroidUniqueID: rowid})
                return fail
            deferred.addErrback(_cleanup)
        config.set("DEFAULT", "key", "kermit-{}".format(rowid))

        workspaces = config.getlist(target, "workspaces")
        server = config.get(target, "server", default="localhost")
        fprints = config.getlist(target, "fingerprints", default=[])

        ds = []
        for ws in workspaces:
            section = config.section(ws, category="workspace")
            shuser = section.get("username", default=user)

            logging.info("Spawning workspace process: {}@{}"
                         .format(shuser, server))
            d = ssh.spawn_process(server, fprints,
                                  shuser, self.get_private_keys(shuser),
                                  "-w", config)

            ds.append(d)

        dl = defer.DeferredList(ds, fireOnOneErrback=True, consumeErrors=True)
        dl.addCallbacks(lambda _: ssh.spawn_process(server, fprints, user,
                                                    self.get_private_keys(user),
                                                    "", config),
                        lambda fail: fail.value.subFailure)
        dl.chainDeferred(deferred)
    

    @command(helphint="<patch> to <target> against <buglist>")
    def commit(self, msg, arg):
        commitdiff, target, buglist = self._parse(arg)

        d = defer.Deferred()
        d.addCallbacks(lambda k: msg.reply("Commit started as {}".format(k)),
                       lambda f: msg.reply("Failed to start commit: {}"
                                           .format(f)))

        self.submit_commit(d, JID(msg.sender_full).user, commitdiff,
                           target, buglist)


    @command(helphint="<commit-id>")
    def abort(self, msg, arg):
        d = defer.Deferred()
        d.addCallbacks(lambda k: msg.reply("Commit {} abort".format(k)),
                       lambda f: msg.reply("Failed to abort commit {}: {}"
                                           .format(arg, f)))
        self.abort_commit(d, int(arg))


    @command(helphint="<commit-id>")
    def resubmit(self, msg, arg):
        d = defer.Deferred()
        d.addCallbacks(lambda k: msg.reply("Commit {} resubmitted".format(k)),
                       lambda f: msg.reply("Failed to resubmit commit {}: {}"
                                           .format(arg, f)))
        self.resubmit_commit(d, int(arg))


    @command(synonyms=(('clear', 'commit'),), hidden=True)
    def clear_commits(self, msg, arg):
        if arg:
            self.db.delete({EndroidUniqueID: int(arg)})
        else:
            self.db.delete()
        self.notify_waiters()
        msg.reply("Done.")
