
import struct
import logging

from twisted.conch import error
from twisted.conch.ssh import transport, keys, connection, channel, common, userauth
from twisted.internet import reactor, defer, protocol

class _TriggerTransport(transport.SSHClientTransport):

    def __init__(self, deferred, fingerprints, username, keys, args, config):
        self._deferred = deferred
        self._fingerprints = fingerprints
        self._username = username
        self._keys = keys
        self._args = args
        self._config = config

    def verifyHostKey(self, key, fingerprint):
        for f in self._fingerprints:
            if f.lower() == 'ignore' or f.lower() == fingerprint.lower():
                return defer.succeed(True)
        logging.warning("Fingerprint doesn't match: " + fingerprint)
        d, self._deferred = self._deferred, None
        err = ValueError("fingerprint ({}) doesn't match config"
                         .format(fingerprint))
        d.errback(err)
        return defer.fail(err)

    def connectionSecure(self):
        self.requestService(_TriggerAuth(str(self._username),
                                         _TriggerConn(self._deferred,
                                                      self._args, self._config,
                                                      self._connection_authed),
                                         self._keys))

    def _connection_authed(self):
        self._deferred = None

    def connectionLost(self, reason):
        logging.warning("Lost connection: " + str(reason))
        d, self._deferred = self._deferred, None
        if d is not None:
            d.errback(ValueError("failed to setup connection for user {}"
                                 .format(self._username)))

class _TriggerAuth(userauth.SSHUserAuthClient):

    def __init__(self, username, conn, keys):
        userauth.SSHUserAuthClient.__init__(self, username, conn)
        self._lastkey = None
        self._keys = keys

    def getPassword(self, prompt=None):
        return

    def getPublicKey(self):
        if not self._keys:
            self._lastkey = None
            return None
        self._lastkey = self._keys.pop(0)
        return self._lastkey.public()

    def getPrivateKey(self):
        return defer.succeed(self._lastkey)

class _TriggerConn(connection.SSHConnection):

    def __init__(self, deferred, args, config, authfn):
        connection.SSHConnection.__init__(self)
        self._deferred = deferred
        self._args = args
        self._config = config
        self._authfn = authfn

    def serviceStarted(self):
        connection.SSHConnection.serviceStarted(self)
        self._authfn()
        self.openChannel(_TriggerChan(self._deferred, self._args, self._config,
                                      conn=self))
        self._deferred = None

class _TriggerChan(channel.SSHChannel):
    name = 'session'

    def __init__(self, deferred, args, config, conn):
        channel.SSHChannel.__init__(self, conn=conn)
        self._deferred = deferred
        self._config = config
        self._command = "kermit_trigger"
        if args:
            self._command += " " + args
        self._retcode = None
        self._sig = None

    def channelOpen(self, data):
        d = self.conn.sendRequest(self, 'exec', common.NS(self._command),
                                  wantReply=True)
        d.addCallbacks(self._spawned, self._failed)

    def _spawned(self, _):
        self._config.write(self)
        self.conn.sendEOF(self)

    def _failed(self, failure):
        d, self._deferred = self._deferred, None
        d.errback(failure)

    def request_exit_status(self, data):
        self._retcode = struct.unpack("!i", data)[0]
        logging.info("Exit status received: {}".format(self._retcode))

    def request_exit_signal(self, data):
        strlen = struct.unpack("!i", data[:4])[0]
        sig, core, msglen = struct.unpack("!{}s?i".format(strlen),
                                          data[4:strlen+9])
        msg, langlen = struct.unpack("!{}si".format(msglen),
                                     data[9+strlen:13+strlen+msglen])
        lang = struct.unpack("!{}s".format(langlen), data[13+strlen+msglen:])[0]
        logging.info("Signal {}. Core: {}. Msg: {}. Lang: {}"
                     .format(sig, core, msg, lang))
        self._retcode = -1
        self._sig = sig

    def dataReceived(self, data):
        logging.info("Received some data: " + repr(data))

    def extReceived(self, dataType, data):
        if dataType == 1:
            logging.info("Stderr data: {}".format(repr(data)))
        else:
            logging.warning("Unexpected ext data (type {}): {}"
                            .format(dataType, repr(data)))

    def closed(self):
        logging.info("Channel has closed")
        d, self._deferred = self._deferred, None
        if self._retcode == 0:
            d.callback(None)
        else:
            d.errback((self._retcode, self._sig))

class _TriggerFactory(protocol.ClientFactory):

    def __init__(self, deferred, fingerprints, username, keys, args, config):
        self._deferred = deferred
        self._fingerprints = fingerprints
        self._username = username
        self._keys = keys
        self._args = args
        self._config = config

    def buildProtocol(self, addr):
        trans = _TriggerTransport(self._deferred, self._fingerprints,
                                  self._username, self._keys, self._args,
                                  self._config)
        self._deferred = None
        return trans

    def clientConnectionFailed(self, connector, reason):
        d, self._deferred = self._deferred, None
        d.errback(reason)

def spawn_process(server, fingerprints, username, keys, args, config, port=22):
    d = defer.Deferred()
    factory = _TriggerFactory(d, fingerprints, username, keys, args, config)
    reactor.connectTCP(server, port, factory)
    return d
