# -----------------------------------------------------------------------------
# EnDroid - Kermit plugin web interface
# Copyright 2014, Ensoft Ltd
# -----------------------------------------------------------------------------

import re
import json
import os.path
import logging
import datetime
import cfgparser

from endroid.database import EndroidUniqueID, LIKE, Value, EndroidLastModified

from twisted.web import resource, server, util
from twisted.python import failure
from twisted.internet import defer, reactor

def _parse_date(dt, fmt="%Y-%m-%dT%H:%M:%S.%fZ"):
    return datetime.datetime.strptime(dt, fmt)


class KermitPage(resource.Resource):
    isLeaf = True

    def __init__(self, plugin):
        resource.Resource.__init__(self)
        self._plugin = plugin

    def _render_success(self, request, key):
        request.setHeader("Content-Type", "text/json")
        request.write(self._plugin.http.get_template("kermit/response.json")
                      .render({
            "success": json.dumps(key),
        }).encode('utf-8'))
        request.finish()
        return server.NOT_DONE_YET

    def _render_fail(self, request, fail):
        request.setHeader("Content-Type", "text/json")
        request.setResponseCode(500)
        request.write(self._plugin.http.get_template("kermit/response.json")
                      .render({
            "fail": str(fail.value),
        }).encode("utf-8"))
        request.finish()
        return server.NOT_DONE_YET


class KermitStager(KermitPage):
    def render_GET(self, request):
        target = request.args['target'][0]
        config = self._plugin.read_config(target)
        request.setHeader("Content-Type", "application/json")
        if not config.has_section(target):
            request.setResponseCode(404)
            return json.dumps({"fail": "no such section {}".format(target)})

        result = []
        for stage in config.getlist(target, "commands"):
            section = config.section(stage, category="command")
            result.append({
                "name": stage,
                "deps": section.getlist("dependencies", default=[]),
            })
        return json.dumps({
            "stages": result,
            "workspaces": config.getlist(target, "workspaces", default=[])
        })


class KermitLister(KermitPage):
    def render_GET(self, request):
        field = request.args['arg'][0]
        rows = self._plugin.db.fetch((field,), distinct=True)
        request.setHeader("Content-Type", "application/json")
        return json.dumps({field: [r[field] for r in rows]})


class KermitUpdater(KermitPage):
    def _error(self, request, error):
        return self._plugin.http.get_template("kermit/response.json").render({
            'fail': error,
        }).encode("utf-8")

    def render_POST(self, request):
        request.setHeader("Content-Type", "application/json")
        m = re.match("/kermit/update/kermit-(\d+)", request.path)
        if not m:
            return self._error(request, "Invalid update URL")

        key = int(m.group(1))
        data = json.loads(request.content.read())

        row = self._plugin.db.fetch(("status",), {EndroidUniqueID: key})
        if not row:
            return self._error(request, "Unknown commit ID: {}".format(key))
        row = row[0]

        olddata = json.loads(row['status'])
        olddata['currstatus'] = data
        self._plugin.db.update({'status': json.dumps(olddata)},
                               {EndroidUniqueID: key})

        self._plugin.notify_waiters()

        return self._plugin.http.get_template("kermit/response.json").render({
            'success': key,
        }).encode('utf-8')


class KermitQuerier(KermitPage):
    def render_POST(self, request):
        def _cancel(_):
            logging.info("Cancelling long poll due to client disconnect")
            self._plugin.waiters.remove((self, request))
        lastupdate = request.args.get('wait', None)
        logging.info("Query received. Args are {}".format(request.args))
        logging.info("Query received. Wait is {}".format(lastupdate))
        if lastupdate:
            res = self._plugin.db.fetch((EndroidLastModified,), max=True)
            logging.info("Latest of all updates was {}".format(res))
            if res:
                latest = datetime.datetime.strptime(res[0].last_modified, "%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%S.%fZ")
                logging.info("Waiting for {}. Latest is {}.".format(lastupdate, latest))
                if latest <= lastupdate[0]:
                    self._plugin.waiters.append((self, request))
                    request.notifyFinish().addErrback(_cancel)
                    return server.NOT_DONE_YET
            
        return self.respond(request)

    def respond(self, request):
        jobs = self._plugin.db.fetch(("status", EndroidLastModified), {
            "user": LIKE(request.args['username'][0]),
            "target": LIKE(request.args['target'][0]),
            EndroidLastModified: Value(request.args['lastmod'][0], "<"),
        })
        request.setHeader("Content-Type", "application/json")
        res = []
        latest = ''
        for job in jobs:
            status = json.loads(job['status'])
            cfg = self._plugin.read_config(status['inputargs']['target'])
            config = {
              "commands": {},
              "workspaces": {},
            }
            for stage in status['currstatus'].get('running', []):
                section = cfg.section(stage, category="command")
                config['commands'][stage] = dict(section.items())
            for stage in status['currstatus'].get('failed', []):
                if cfg.has_section(stage, category='command'):
                    section = cfg.section(stage, category="command")
                    config['commands'][stage] = dict(section.items())
            for stage in status['currstatus'].get('stages', []):
                section = cfg.section(stage, category="command")
                config['commands'][stage] = dict(section.items())
            for ws in cfg.getlist(status['inputargs']['target'], "workspaces"):
                section = cfg.section(ws, category="workspace")
                config['workspaces'][ws] = dict(section.items())
            lastmod = datetime.datetime.strptime(job.last_modified, "%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            res.append({
                "status": status,
                "config": config,
                "key": "kermit-{}".format(job.id),
                "lastmod": lastmod,
                "status_string":
                    "Completed" if status['currstatus'].get('completed_at',
                                                            False) else
                    "Failed" if status['currstatus'].get('failed_at', False) or
                                status['currstatus'].get('check_failed',
                                                         False) else
                    "Aborted" if status.get('aborted_at', False) else
                    "Running" if status['currstatus'].get('running', False) else
                    "Queued",
            })
            if lastmod > latest:
                latest = lastmod
        request.write(json.dumps({"last_update": latest, "jobs": res}))
        request.finish()
        return server.NOT_DONE_YET


class KermitSubmitter(KermitPage):
    def render_POST(self, request):
        d = defer.Deferred()
        d.addCallbacks(lambda k: self._render_success(request, k),
                       lambda f: self._render_fail(request, f))

        try:
            if "resubmitKey" in request.args:
                m = re.match("kermit-(\d+)", request.args["resubmitKey"][0])
                if not m:
                    raise ValueError("Malformed entry index ({})"
                                     .format(request.args["resubmitKey"][0]))
                key = int(m.group(1))
                self._plugin.resubmit_commit(d, key)
            else:
                self._plugin.submit_commit(d,
                                       request.args['username'][0],
                                       request.args['commitdiff'][0],
                                       request.args['target'][0],
                                       request.args['buglist'][0],
                                       request.args.get('commitmsg', [''])[0],
                                       request.args['stages'][0],
                                       ws_disable=request.args['ws_disable'][0],
                                       ws_reset=request.args['ws_reset'][0])
        except Exception as e:
            logging.exception(e)
            d.errback(e)

        return server.NOT_DONE_YET


class KermitAborter(KermitPage):
    def render_POST(self, request):
        d = defer.Deferred()
        d.addCallbacks(lambda k: self._render_success(request, k),
                       lambda f: self._render_fail(request, f))
        try:
            m = re.match("kermit-(\d+)", request.args["key"][0])
            if not m:
                raise ValueError("Malformed entry index ({})"
                                 .format(request.args["key"][0]))
            key = int(m.group(1))
            self._plugin.abort_commit(d, key)
        except Exception as e:
            logging.exception(e)
            d.errback(e)

        return server.NOT_DONE_YET


class KermitQueue(KermitPage):
    def render_GET(self, request):
        items = self._plugin.db.fetch((EndroidUniqueID, "user",
                                       "target", "status"))

        for item in items:
            item['key'] = "kermit-{}".format(item[EndroidUniqueID])
            item['status'] = json.loads(item['status'])
            config = self._plugin.read_config(item['target'])
            item['config'] = config

        return self._plugin.http.get_template("kermit/queue.html").render({
            "items": items,
            "username": request.getUser(),
        }).encode("utf-8")


class KermitProfile(KermitPage):
    def render_GET(self, request):
        return self._plugin.http.get_template("kermit/profile.html").render({
            "username": request.getUser(),
            "keys": self._plugin.list_keys(request.getUser()),
        }).encode("utf-8")


class KermitUploadKey(KermitPage):
    def render_POST(self, request):
        try:
            key = self._plugin.add_key(request.getUser(),
                                       request.args["privatekey"][0])
        except Exception as e:
            return self._render_fail(request, failure.Failure(e))
        else:
            return self._render_success(request, key)


class KermitGenKey(KermitPage):
    def render_GET(self, request):
        def _save_key(key):
            return self._plugin.add_key(request.getUser(),
                                        key.toString("OPENSSH"))
        d = self._plugin.generate_key(request.getUser())
        d.addCallback(_save_key)
        d.addCallbacks(lambda k: self._render_success(request, k),
                       lambda f: self._render_fail(request, f))
        return server.NOT_DONE_YET

class KermitDelKey(KermitPage):
    def render_GET(self, request):
        try:
            self._plugin.remove_key(request.getUser(), request.args['idx'][0])
        except Exception as e:
            return self._render_fail(request, failure.Failure(e))
        else:
            return self._render_success(request, 'done')


class KermitRoot(resource.Resource):
    def __init__(self, plugin):
        resource.Resource.__init__(self)
        self._plugin = plugin
        self.putChild("", plugin.http.authed_resource(KermitQueue(plugin)))
        self.putChild("update", KermitUpdater(plugin))
        self.putChild("submit", KermitSubmitter(plugin))
        self.putChild("list", KermitLister(plugin))
        self.putChild("stages", KermitStager(plugin))
        self.putChild("query",
                      plugin.http.authed_resource(KermitQuerier(plugin)))
        self.putChild("abort", KermitAborter(plugin))
        self.putChild("profile",
                      plugin.http.authed_resource(KermitProfile(plugin)))
        self.putChild("uploadkey",
                      plugin.http.authed_resource(KermitUploadKey(plugin)))
        self.putChild("generatekey",
                      plugin.http.authed_resource(KermitGenKey(plugin)))
        self.putChild("removekey",
                      plugin.http.authed_resource(KermitDelKey(plugin)))

    def render_GET(self, request):
        return util.redirectTo(request.path + '/', request)
